/*
 *				**** Linked List ****
 *
 * This file contains routines that handle linked lists
 *
 * This linked list will be kept sorted by key
 * with the smallest key at the front (head) of the list
 * and the largest key at the back (tail)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "datademo.h"

struct LinkedList LinkedList; /* linked list */

/* function prototypes */
void add_linked_list(void);
void delete_linked_list(void);
void list_linked_list(void);
void search_linked_list(void);
void display_linked_list(int ent);
int find_free_linked_list(void);
int find_linked_list(int key);
int find_insert_linked_list(int key);

void
init_linked_list(void)
{
	int i;

   /*
    * Initialize Linked List
    *
    * Initialize Head to indicate empty list
    * Value of -1 indicates end of list
    */
    for (i=0; i < MAX_ENTS; i++)
    { /* loop through entries */
		LinkedList.list[i].next = -1;	/* nothing follows */
      LinkedList.list[i].entry.use = USE_NOT;
      LinkedList.list[i].entry.key = 0;		/* no key */
      LinkedList.list[i].entry.desc[0] = '\0';
    }
    LinkedList.head = -1;		/* list is empty */
}

int
do_linked_list(void)
{
	char buf[BUFSIZ];

   while (1)
   {
   	printf("\nOptions are:\n");
      printf("A. Add entries to linked list\n");
      printf("D. Delete entries from linked list\n");
      printf("L. List linked list\n");
      printf("S. Search linked list\n");
      printf("M. Return to Main Menu\n");
      printf("Q. Quit\n");
      printf("Enter option: ");
      gets(buf);
      switch(buf[0])
      {
      case 'a':
      case 'A':
      	add_linked_list();
         break;
      case 'd':
      case 'D':
      	delete_linked_list();
         break;
      case 'l':
      case 'L':
      	list_linked_list();
         break;
      case 's':
      case 'S':
      	search_linked_list();
         break;
      case 'm':
      case 'M':
      	return 1;	/* keep running */

      case 'q':
      case 'Q':
      	return 0;	/* stop running */
         
      default:
      	printf("Invalid option - try again!\n");
      }
   }
   return 0;	/* should never happen */
}

void
add_linked_list(void)
{ /* add an entry to the linked_list */
	char buf[BUFSIZ];
   int key, ent, prev, next;

   /* make sure there is a free link */
   ent = find_free_linked_list();
   if (ent < 0)
   {
   	printf("List is full.  Delete some entries\n");
      return;
   }
   printf("Entry %d is available\n", ent);

   /* ready to add to end of linked_list */
   printf("Enter key: ");
   gets(buf);
   key = atoi(buf);		/* set key */
   do
   {
	   printf("Enter description: ");
	   gets(buf);
	   if (strlen(buf) > MAX_LEN)
      {
      	printf("Name too long.  Try again? (y or n): ");
         gets(buf);
         if (buf[0] == 'n') return;
      }
   } while(strlen(buf) >= MAX_LEN);

   /* set up the new link entry */
   LinkedList.list[ent].entry.key = key;
   LinkedList.list[ent].entry.use = USE_YES;
   strcpy(LinkedList.list[ent].entry.desc, buf);

   /* find where to insert this entry */
   /* insert new link after insert_loc */
   prev = find_insert_linked_list(key);

   /*
    * There are two cases when inserting a list
    * The new entry is added after some existing
    * entry on the list, or it is added at the head
    */
   if (prev < 0)
   { /* insert at head */
   	printf("Adding new entry at start of list\n");
   	/* next entry is the current head of the list */
   	LinkedList.list[ent].next = LinkedList.head;
      LinkedList.list[ent].prev = -1; /* nothing before */
      /* point old head back to new head */
      if (LinkedList.head >= 0)
      { /* make sure there is a head already */
      	LinkedList.list[LinkedList.head].prev = ent;
      }
      /* place new entry at the head of the list */
      LinkedList.head = ent;	/* new head */
		return;
   }

   /*
    * add after insert_loc
    *
    * entry after new link is where insert_loc used
    * to point
    *
    * insert_loc should now point to new entry
    */
   next = LinkedList.list[prev].next;
   printf("Adding new entry between %d and %d\n", prev, next);
   display_linked_list(prev);
   if (next >= 0) display_linked_list(next);

   /* insert ent between prev and next */
   LinkedList.list[ent].next = next;
   LinkedList.list[ent].prev = prev;
   /* update prev & next */
   LinkedList.list[prev].next = ent;
   if (next >= 0)
   { /* not end of list, update next */
   	LinkedList.list[next].prev = ent;
   }
}

void
delete_linked_list(void)
{
	char buf[BUFSIZ];
   int ent, key, next, prev;

   printf("Enter key to delete: ");
   gets(buf);
   key = atoi(buf);
	ent = find_linked_list(key);
   if (ent < 0)
   {
   	printf("Entry for key %d not found\n", key);
      return;
   }

   /* found entry - check with user */
   while (ent >= 0 && LinkedList.list[ent].entry.key == key)
   { /* ask about all entries with the same key */
	   printf("Entry to be removed:\n");
	   display_linked_list(ent);
	   printf("OK? (y or n): ");
	   gets(buf);
	   if (buf[0] != 'y') return;		/* don't remove */

	   /* remove entry from list */
	   prev = LinkedList.list[ent].prev;
	   next = LinkedList.list[ent].next;
	   printf("Removing entry from between %d and %d\n", prev, next);
	   if (prev >= 0) display_linked_list(prev);
	   if (next >= 0) display_linked_list(next);

	   /* patch list around entry to be deleted */
	   LinkedList.list[prev].next = next;
	   LinkedList.list[next].prev = prev;

	   /* entry now deleted, mark it that way */
	   LinkedList.list[ent].entry.use = USE_NOT;

      /* maybe delete next entry? */
      ent = next;
   }
}

void
list_linked_list(void)
{
	int ent, num_ents=0;

   for (ent=LinkedList.head; ent >= 0; ent = LinkedList.list[ent].next)
   {
   	display_linked_list(ent);
   	num_ents++;
   }

   printf("%d entries in table\n", num_ents);
}

void
display_linked_list(int ent)
{
	printf("Entry %d, prev: %d, next: %d, key %d, desc: %s\n",
   	ent, LinkedList.list[ent].prev,
      LinkedList.list[ent].next,
      LinkedList.list[ent].entry.key,
      LinkedList.list[ent].entry.desc);
}

void
search_linked_list(void)
{
	char buf[BUFSIZ];
   int key, ent, num_matches=0;

   printf("Enter key to search for: ");
   gets(buf);
   key = atoi(buf);

   printf("Searching for key: %d\n", key);

   ent = find_linked_list(key);
	if (ent < 0)
   {
   	printf("No entries found for key %d\n", key);
      return;
   }

   while (ent >= 0 && LinkedList.list[ent].entry.key == key)
   { /* find all matches, in case there are duplicate keys */
   	display_linked_list(ent);
      num_matches++;
      ent = LinkedList.list[ent].next;	/* next entry */
   }

   printf("%d matches found\n", num_matches);
}

int
find_free_linked_list(void)
{ /* find a free entry */
	int i;

   for (i=0; i < MAX_ENTS; i++)
   { /* find an empty or deleted entry */
   	if (LinkedList.list[i].entry.use == USE_YES) continue;
      return i;	/* free entry found */
   }
   return -1;		/* no free entry found */
}

int
find_linked_list(int key)
{ /* find entry matching key */
	int i;

   for (i=LinkedList.head; i >= 0; i = LinkedList.list[i].next)
   {
   	if (LinkedList.list[i].entry.key == key) return i;
   }

   return -1;		/* no match was found */
}

int
find_insert_linked_list(int key)
{ /* find place to insert key */
	int ent, next;

   /* get first entry in list */
   ent = LinkedList.head;
   /* if list is empty, return -1 */
   if (ent < 0) return -1;	/* add to front of list */
   /* if key smaller than first (head) entry, return -1 */
   if (key < LinkedList.list[ent].entry.key) return -1;

	for (; ent >= 0; ent = next)
   { /* loop through the list */
   	next = LinkedList.list[ent].next;	/* next entry */
      /* if no next entry, then add after this one */
      if (next < 0)
      {
      	printf("Add entry to end of list\n");
	      return ent;
      }
		/* see if new entry should be inserted before next one */
      if (LinkedList.list[next].entry.key > key)
      {
      	printf("Adding entry before entry %d (key %d)\n",
         	next, LinkedList.list[next].entry.key);
      	return ent;
      }
   }

   /* we should never get here */
   printf("Error while searching linked list\n");
   return -1;	/* do something -- maybe add at front */
}

